package ddejonge.nb3;


import java.util.Arrays;
import java.util.HashMap;
import java.util.Random;

import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.SwingUtilities;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

/**
 * Tool for debugging.
 * Displays property-value pairs in a window.<p>
 * 
 * Each property must first be added with a line number:<br/>
 *  <code>dataMonitor.setProperty(1, "age"); </code><br/>
 * or without a line number:<br/>
 *   <code>dataMonitor.setProperty("age"); </code><p>

 * Then you can set the value of the property:<br/>
 *   <code>dataMonior.setValue("age", 27); </code><br/>
 * <br/>
 * The size of the window can be set by calling:<br/>
 *   <code>dataMonitor.setBounds(x, y, width, height) </code>
 * 
 * @author Dave de Jonge
 *
 */
public class DataMonitor{

	private JFrame frame;
	private JTextPane textPane;
	
	
	private HashMap<String, Object> properties = new HashMap<>();
	private String[] line2prop = new String[1]; //maps each line number to its corresponding property.
	
	private int numLines=0;
	
	
	int width = 500;
	int height = 500;
	int x = 200;
	int y = 200;


	private boolean enabled = false;
	
	//creates a disabled monitor
	public DataMonitor(){
	}
	
	//creates an enabled monitor
	public DataMonitor(String title){
		enable(title);
	}
	
	public void setSize(int width, int height){
		this.width = width;
		this.height = height;
	}
	
	public void enable(String title){
		
		//The contentPane has a JScrollPane, which has a JTextPane
		if(frame == null){
			textPane = new JTextPane();
			JScrollPane scrollPane = new JScrollPane(textPane);						
			
			frame = new JFrame();
			frame.getContentPane().add(scrollPane);
		
			Random random = new Random();
			x += random.nextInt(800);
			
			frame.setBounds(x, y, width, height);
		}
		
		this.enabled = true;
		frame.setTitle(title);
		frame.setVisible(true);
	}
	
	public void disable(){
		
		this.enabled = false;
		if(frame != null){
			frame.setVisible(false);
		}
		
	}
	
	private void updateTextPane(final String text) {
		  SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
		      Document doc = textPane.getDocument();
		      try {
		    	  doc.remove(0, doc.getLength());
		    	  doc.insertString(doc.getLength(), text, null);			        
		      } catch (BadLocationException e) {
		        throw new RuntimeException(e);
		      }
		      /*textPane.setCaretPosition(doc.getLength() - 1);*/
		    }
		  });
	}
	
	
	public void setProperty(String property){
		setProperty(numLines+1,property);
	}
	
	
	public void setProperty(int lineNumber, String property){
		
		if(lineNumber >= line2prop.length){
			resizeArray(lineNumber + 1);
		}
		
		if(lineNumber > numLines){
			numLines = lineNumber;
		}
		
		line2prop[lineNumber] = property;
		properties.put(property, "");
	}
	
	/**
	 * Displays the new value of the property in the window.
	 * 
	 * @param property the name of the property to display.
	 * @param value the value of the property to display.
	 */
	public void setValue(String property, Object value){
		setValue(property, value, true);
	}
	
	/**
	 * Sets the new value of the property in the window. If update is set to 'false' this value is not displayed in the window
	 * until the method update() is called.
	 * 
	 * @param property the name of the property to display.
	 * @param value the value of the property to display.
	 * @param update the window will be updated with the new value immediately if set to 'true'. Otherwise, the new value will not be displayed in the window until the method update() is called. 
	 */
	public void setValue(String property, Object value, boolean update){
		
		if(!this.enabled){
			return;
		}
		
		if(properties.get(property) == null){
			
			throw new IllegalArgumentException("DataMonitor.setValue() Error! Unknown property: " + property);
			
		}else{
			
			//the property already existed, set the value only.
			properties.put(property, value);
		}
		
		if(update){
			update();
		}
		
	}

	public void setValues(HashMap<String, Object> hashMap){
		
		if(!this.enabled){
			return;
		}
		
		for(String key : hashMap.keySet()){
			this.properties.put(key, hashMap.get(key));
		}
		
		update();
	}
	
	
	/**
	 * Makes sure that the current values assigned to the properties are displayed. 
	 * 	
	 */
	public void update(){
		
		if(!this.enabled){
			return;
		}
		
		String text = "";
		for(int i=0; i<line2prop.length; i++){
			
			if(line2prop[i] != null){
				text+= line2prop[i] + " " + properties.get(line2prop[i]);
			}	
			
			text += "\n";
		}
		
		
		updateTextPane(text);
	}
		
	public void remove(String key){
		properties.remove(key);
	}
	
	public void clear(){
		Arrays.fill(line2prop, null);
		properties.clear();
	}
	
	private void resizeArray(int newSize){
		
		String[] newContent = new String[newSize];
		
		for(int i=0; i<line2prop.length; i++){
			newContent[i] = line2prop[i];
		}
		
		line2prop = newContent;
	}
	
	public boolean isEnabled(){
		return this.enabled;
	}
}
