package ddejonge.nb3;

import java.util.Iterator;





/**
 * Class that represents a set of agents.
 * Each agent is supposed to have an ID which is an integer between 0 and 63.
 * 
 * This class is just a wrapper around a long int. If the i-th bit of this long is set to 1 this represents that the agent with
 * ID i is in the set, if the i-th bit of this long is set to 0 it represents that the agent with ID i is not in the set.
 * 
 * NOTE: The fact that we are using a long int to represent a set of agents means that NB3 cannot handle negotiation domains with more
 * than 64 agents.
 * 
 * @author Dave de Jonge
 *
 */
public class NB3AgentSet implements Iterable<Integer>{
	
	final public static int MAX_SIZE = 64;
	
	private long setId;
	
	
	//CONSTRUCTORS
	
	/**
	 * Creates the empty set.
	 */
	public NB3AgentSet(){
		this.setId = 0;
	}
	
	/**
	 * Creates a set based on the given id.
	 */
	NB3AgentSet(long setId){
		this.setId = setId;
	}
	
	/**
	 * Copy constructor
	 * @param original
	 */
	public NB3AgentSet(NB3AgentSet original){
		this.setId = original.getId();
	}
	
	
	//GETTERS AND SETTERS
	long getId(){
		return this.setId;
	}
	
	/**
	 * Adds the first n elements.
	 * @param size the number of elements to put in the set.
	 */
	void fill(int size){
		this.setId = getFilledSet(size);
	}

	public void add(int agentID){
		this.setId = addElement(setId, agentID);
	}
	
	public boolean contains(int agentID){
		return containsElement(setId, agentID);
	}
	
	public void remove(int agentID){
		this.setId = removeElement(setId, agentID);
	}
	
	void addAll(long otherSetId){
		this.setId = getUnion(setId, otherSetId);
	}
	
	public void addAll(NB3AgentSet otherSet){
		this.setId = getUnion(setId, otherSet.setId);
	}
	
	boolean containsAll(long otherSetId){
		return containsSubset(setId, otherSetId);
	}
	
	public boolean containsAll(NB3AgentSet otherSet){
		return containsSubset(setId, otherSet.setId);
	}
	
	void removeAll(long otherSetId){
		this.setId = NB3AgentSet.removeSet(this.setId, otherSetId);
	}
	
	public void removeAll(NB3AgentSet otherSet){
		this.setId = NB3AgentSet.removeSet(this.setId, otherSet.getId());
	}
	
	
	public void clear(){
		this.setId = 0;
	}
	
	public boolean isEmpty(){
		return this.setId == 0;
	}
	
	/**
	 * Removes all elements of this set and adds all elements of the given set.
	 * 
	 * @param other
	 */
	public void makeEqualTo(NB3AgentSet otherSet){
		this.setId = otherSet.setId;
	}
	
	
	void makeEqualTo(long id){
		this.setId = id;
	}
	
	
	public int size(){
		
		int size = 0;
		for(int i=0; i<NB3AgentSet.MAX_SIZE; i++){
			if(containsElement(this.setId, i)){
				size++;
			}
		}
		
		return size;
	}
	
	public String toString(){
		String s = "{";
		for(int i=0; i<NB3AgentSet.MAX_SIZE; i++){
			if(containsElement(this.setId, i)){
				s += "" + i + ", ";
			}
		}
		
		//remove the last comma.
		if(s.length() > 1){
			s = s.substring(0, s.length() - 2);
		}
		
		s+= "}";
		
		return s;
	}
	
	
	//METHODS THAT CREATE NEW SETS.
	
	/**
	 * Makes a copy of this set.
	 * @return
	 */
	public NB3AgentSet copy(){
		return new NB3AgentSet(this.setId);
	}
	
	/**
	 * Creates a new AgentSet that contains all agents in this set that are not in the given other set.
	 * @param otherSet
	 * @return A new set.
	 */
	public NB3AgentSet subtract(NB3AgentSet otherSet){
		long newSetId = NB3AgentSet.removeSet(this.setId, otherSet.setId);
		return new NB3AgentSet(newSetId);
	}
	

	
	
		
	
	//STATIC METHODS.
	
	//Anything that returns a long is made private, because the user shouldn't know about the 
	//underlying implementation. 
	
	private static long addElement(long set, int elementId){
		return set | (1L<<elementId);
	}
	
	private static boolean containsElement(long set, int elementId){
		return (1L<<elementId & set) == (1L<<elementId);
	}
		
	private static long removeElement(long set, int elementId){
		return set & ~(1L<<elementId);
	}
	
	
	private static long getEmptySet(){
		return 0;
	}
	
	/**
	 * Returns the id of a set that contains all first n elements.
	 * @param size the number of elements to put in the set.
	 * @return
	 */
	private static long getFilledSet(int size){
		if(size >= MAX_SIZE){
			return ~(0L);
		}else{
			return (1L<<size) - 1;
		}
	}
	
	private static long getUnion(long set1, long set2){
		return set1 | set2;
	}
	
	private static long getIntersection(long set1, long set2){
		return set1 & set2;
	}
	
	private static boolean containsSubset(long set, long subset){
		return (subset & set) == subset;
	}
	
	private static long removeSet(long set, long subset){
		return set & ~subset;
	}

	
	
	@Override
	public Iterator<Integer> iterator() {
		return new AgentSetIterator(this.getId());
	}
	
	
	class AgentSetIterator implements Iterator<Integer>{

		private int currentIndex = 0;  //if this value is n, then Next() will return the n'th element
		private int nextElement = -1;  //the next value m for which container is true.
		private long setId;
		
		AgentSetIterator(long setId){
			this.setId = setId;
		}
		
		@Override
		public boolean hasNext() {
			
			for(int i=currentIndex; i<MAX_SIZE; i++){
				if(NB3AgentSet.containsElement(setId, i)){
					nextElement = i;
					return true;
				}
			}
			nextElement = -1;
			return false;
		}

		@Override
		public Integer next() {
			
			if(nextElement != -1 && nextElement >= currentIndex){
				currentIndex = nextElement+1;
				return nextElement;
			}
			
			//else:
			//either there is no next element, (nextElement == -1)
			//or hasNext() hasn't been called so nextElement is not updated (nextElement < currentIndex)
			
			if(hasNext()){
				currentIndex = nextElement+1;
				return nextElement;
			}
			
			return -1;
		}

		
		@Override
		public void remove() {
			throw new RuntimeException("NB3AgentSet.AgentSetIterator.remove() Error! remove() is not implemented");
			
		}

		
	}

}
