package ddejonge.nb3;



import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.PriorityQueue;

import ddejonge.nb3.NB3Message.MsgType;



/**
 * Base class for the implementation of the NB3 algorithm.
 * @author Dave de Jonge
 *
 */
public abstract class NB3Algorithm {

	//FIELDS
	final int NUM_NODE_TYPES;
	
	/**
	 * Logger to record information about the search algorithm.
	 * (e.g. for each node it records all the bounds for all the negotiators.)
	 */
	NB3Logger search_logger = new NB3Logger();
	
	/**
	 * Logger to record information about  the negotiation algorithms.
	 * (e.g. the proposals received and the proposals made)
	 */
	NB3Logger nego_logger = new NB3Logger();
	
	/** Path to the folder where the log files will be stored.*/
	private String logFolderPath = null;
	
	/** Tool for displaying debug information in a window at run-time. */
	NB3DataMonitor dataMonitor = new NB3DataMonitor(this);
	
	/** The world state at the beginning of the negotiation session (does not change during the negotiations).*/
	NB3WorldState initialState; 
	
	/** The current world state (may change during the negotiations).*/
	NB3WorldState currentState; 
	
	/** The ID of the agent on which this algorithm is running.*/
	int myAgentID;
	
	/** The name of the agent on which this algorithm is running. */
	String myAgentName = "";
	
	/**
	 * This field holds information about the negotiation domain, such as the names of the agents, the time
	 * that negotiations started, and the deadline of the negotiations. 
	 */
	NB3SessionInfo sessionInfo;
	
	/** Stores all proposals: rejected, accepted,  found by us, proposed to us and confirmed */
	HashMap<String, NB3Proposal> allProposals = new  HashMap<String, NB3Proposal>(30*1000); 
		
	/** The object that represents the search tree. */
	protected NB3Tree theTree = null;
	
	
	/**
	 * All rational deals found by the search algorithm, ordered by selfishness.
	 */
	PriorityQueue<NB3Proposal> foundByUs = new PriorityQueue<NB3Proposal>(1000); 	
	
	
	protected LinkedList<NB3Proposal> proposedToUs = new LinkedList<NB3Proposal>();   		//deals proposed to us
	protected LinkedList<NB3Proposal> proposedByUs = new LinkedList<NB3Proposal>();   		//deals proposed by us
	protected ArrayList<NB3Proposal> confirmed  = new ArrayList<NB3Proposal>(100);    		//confirmed proposals
	protected ArrayList<NB3Proposal> disccarded  = new ArrayList<NB3Proposal>(100); 		//deals discarded for not being altruistic enough.
	
	
	/** The aspiration level for the agent.*/
	double my_asp;							
	
	/**the minimum value for my satisfaction for being acceptable.*/
	double my_deadline_level = 0.01;
	
	/**concession degree, the higher this value, the faster it concedes.*/
	double my_concession_degree = 2;   
	
	
	/**the aspiration level for the opponents.*/
	double opp_asp; 			
	
	/**the value we want opp_asp to have at the deadline.*/
	double opp_deadline_level = 0.99;		
	
	/**concession degree, the higher this value, the faster it concedes.*/
	double opp_concession_degree = 4;   	
	
	

	
	
	int numNodes = 0; //total number of nodes in the tree, also used to create an id for each new node
	 //this cannot be a field of the tree, because it would be reset each time a deal
	 //is confirmed.
	
	protected int typeOfNodeToSplit = -1; //represents the type of node to split. The root node has type 0 by default.
	//0=Sender, 1=CityNode, 2=ReceiverNode
	
	int numFoundByUs = 0;   	
	
	/**The total number of rational deals proposed to us by others. */
	int numProposedToUs  = 0;	
	
	/**The number of deals we have proposed. */
	int numProposedByUs = 0; 
	
	int numAcceptedByUs = 0;
	
	//The number of confirmed deals we are committed to. (i.e. the number of confirmed deals in which we are participating).
	int numCommitments = 0;
	
	//For debugging purposes:
	/*MyTreeViewer mtv = null;*/ 	
	String status = "--";

	

	
	
	
	/**
	 * Returns the number of rational deals found by the search algorithm. 
	 */
	public int getNumFoundByUs() {
		return numFoundByUs;
	}

	/**
	 * Returns the number of deals that have been proposed to this agent.
	 * @return
	 */
	public int getNumProposedToUs() {
		return numProposedToUs;
	}

	/**
	 * Returns the number of deals that have been proposed by this agent.
	 * @return
	 */
	public int getNumProposedByUs() {
		return numProposedByUs;
	}
	
	/**
	 * Returns the number of deals that have been accepted by this agent.
	 * @return
	 */
	public int getNumAcceptedByUs() {
		return numAcceptedByUs;
	}
	
	
	/**
	 * Returns the number of deals that have been confirmed.
	 * @return
	 */
	public int getNumDealsMade(){
		return numCommitments;
	}

	
	protected NB3WorldState getCurrentState(){
		return this.currentState;
	}
	
	protected String[] getAgentNames(){
		return this.sessionInfo.agentNames;
	}
	
	protected int getAgentID(String name){
		return this.sessionInfo.getAgentID(name);
	}
	
	
	//CONSTRUCTORS
	
	/**
	 * Constructs an algorithm object with only 1 type of node (apart from the root).
	 * 
	 */
	public NB3Algorithm(){
		NUM_NODE_TYPES = 1;
	}
	
	/**
	 * Constructs an algorithm object with the given number of node types.
	 * 
	 * @param numNodeTypes The number of types of nodes, excluding the root node.
	 */
	public NB3Algorithm(int numNodeTypes){
		NUM_NODE_TYPES = numNodeTypes;
	}
	
	
	
	
	
	public void enableLoggers(String folderPath){
		
		this.logFolderPath = folderPath;
		
		//if the algorithm has already been initialized then we can enable the loggers directly.
		if(this.sessionInfo != null){
			search_logger.enable(sessionInfo, logFolderPath, this.myAgentName + "_search.log", false);
			nego_logger.enable(sessionInfo, logFolderPath,  this.myAgentName + "_nego.log", false);
		}
		
		//otherwise they will be enabled once initialize() is called.
		

	}
	
	public void enableMonitor(boolean enable){
		if(enable){
			this.dataMonitor.enable(this.myAgentName);
		}else{
			this.dataMonitor.disable();
		}
	}
	
	
	/**
	 * Sets the speed of concession of the algorithm.
	 * @param opp_concession_degree
	 * @param my_concession_degree
	 */
	public void setConcessionDegrees(double opp_concession_degree, double my_concession_degree){
		this.opp_concession_degree = opp_concession_degree;
		this.my_concession_degree = my_concession_degree;
	}
	
	
	
	
	//INITIALIZERS
	
	/**
	 * Once all information necessary for the negotiation algorithm is available, the agent should call this method to start the algorithm. 
	 * 
	 * @param myName The name of the agent.
	 * @param agentNames The name of all agents in the negotiation session, including the name of this agent.
	 * @param initialState The WorldState object that represents the state at the beginning of the negotiations.
	 * @param deadline The deadline after which the negotiations will stop.
	 */
	public void initialize(String myName, String[] agentNames, NB3WorldState initialState, long deadline){
		initialize(myName, agentNames, initialState, deadline, new NB3NodeQueue(NUM_NODE_TYPES + 1)); //add 1 because the root node is not counted in NUM_NODE_TYPES.
	}
	
	/*
	void initialize(String myName, String[] agentNames, NB3WorldState initialState, long deadline, NB3NodeQueue queue){
		initialize(myName, agentNames, initialState, deadline, queue);
	}*/
	
	/**
	 * Creates a new Tree object, calculates the bounds of the root node,
	 * calculates the glb and gub, and sets the initial offer values.
	 *  
	 * @param agentNumber the id that we use for the agent that owns this algorithm.
	 * @param queue The queue to store the nodes, ordered by their expansion heuristic
	 * @param rootLabel the label of the root node
	 */
	void initialize(String myName, String[] agentNames, NB3WorldState initialState, long deadline, NB3NodeQueue queue){
		
		
		
		//1. CHECK IF THE LIST OF AGENT NAMES DOES NOT CONTAIN DUPLICATE NAMES, AND CHECK THAT MY NAME IS IN IT.
		boolean containsMyName = false;
		for(int i=0; i<agentNames.length; i++){
			
			if(agentNames[i].equals(myName)){
				containsMyName = true;
				this.myAgentID  = i;
			}
			
			for(int j=0; j<agentNames.length; j++){
				
				if(i == j){
					continue;
				}
				
				if(agentNames[i].equals(agentNames[j])){
					throw new IllegalArgumentException("Nb3Algorithm.initialize() Error! set of agent names contains duplicate name " + agentNames[i]);
				}
				
			}
		}		
		if(!containsMyName){
			throw new IllegalArgumentException("Nb3Algorithm.initialize() Error! list of agent names does not contain my name (" + myName + ")");
		}

		
		this.sessionInfo = new NB3SessionInfo(myName, agentNames, deadline);
		this.myAgentName = myName;

		if(dataMonitor.isEnabled()){
			//if the datamonitor is enabled, then make sure its title is set to the agent name.
			dataMonitor.enable(myAgentName);
		}
		
		
		this.initialState = initialState; 
		this.currentState = initialState.copy();
		
		this.sessionInfo.startTime = System.currentTimeMillis();
		
		if(theTree == null){
			theTree = new NB3Tree(this, queue, sessionInfo.numAgents);
		}else{
			theTree.clear();
		}
		//Set the upper- lower-, and intermediate bounds of the root node
		NB3AgentSet emptyAgentSet = new NB3AgentSet();
		ArrayList<NB3Node> emptyBranch = new ArrayList<NB3Node>(0);
		this.calculateBounds(theTree.root, emptyBranch, emptyAgentSet); //do not pass null, instead pass an empty arraylist.
		
		//set the initial offer values equal to the global upper bounds.
		for(int ag=0; ag<sessionInfo.numAgents; ag++){
			theTree.offer[ag] = theTree.getGUB(ag);
		}
		
		//Calculate the expansion heuristic of the root node.
		float eh = this.calculateExpansionHeuristic(theTree.root);
		theTree.root.setExpansionHeuristic(eh);
		
		theTree.original_gub = new float[sessionInfo.numAgents];
		for(int j=0; j<sessionInfo.numAgents; j++){
			theTree.original_gub[j] = theTree.getGUB(j);
		}
		
		theTree.original_rv = new float[sessionInfo.numAgents];
		for(int j=0; j<sessionInfo.numAgents; j++){
			theTree.original_rv[j] = theTree.getRV(j);
		}
		
		if(logFolderPath != null){
			search_logger.enable(sessionInfo, logFolderPath, this.myAgentName + "_search.log", false);
			nego_logger.enable(sessionInfo, logFolderPath,  this.myAgentName + "_nego.log", false);
			
			search_logger.logNode(theTree.root);
			search_logger.logBounds(theTree.root, emptyAgentSet);
			
		}
		
	}
	
	
	
	
	//METHODS
		
	
	//***PART 1: EXPAND
	
	/**
	 * Picks the best unexplored node in the tree and adds children to it.
	 */
	public void expand(){		
		
		//get the node with highest expansion heuristic	
		NB3Node nodeToSplit = chooseNode();							   	
		
		//get the labels we want to assign to the children of that node.
		List<Object> splitLabels = getSplitLabels(nodeToSplit);   		

		//create the new child nodes and calculate their values.
		splitNode(nodeToSplit, splitLabels);						   		
			
		
		
		//FOR DEBUGGING:
		
		//Update the data displayed in the data monitor. 
		this.dataMonitor.setValues();
		
		//Update the log file.
		if(nodeToSplit != null){
			this.search_logger.logln("Nb3Algorithm.expand() Number of rational deals found: " + getNumFoundByUs());
		}

	}
	
	
	
	private NB3Node chooseNode(){
		
		//get the next type of node that we want to split.
		typeOfNodeToSplit = getNextTypeOfNodeToSplit(typeOfNodeToSplit); 
		
		
		//pick the next node to expand.
		NB3Node nextNode;
		do{
			
			nextNode = theTree.queue.poll(typeOfNodeToSplit);	
			
			if(nextNode == null){
				return null;
			}
			
			//repeat if the node has been pruned (a node A is pruned when a proposal corresponding to another node B is accepted and A is not in the subtree under B.) 
		}while(nextNode.isPruned);
		
		
		// Write info about the chosen node to the log file.
		search_logger.logChosenNode("Nb3Algorithm.chooseNode()", nextNode);
		
		return nextNode;		
	}
	
	
	protected NB3Node getRootNode(){
		return this.theTree.root;
	}
	
	
	
	//NOTE: THE TWO METHODS BELOW HAVE DIFFERENT PURPOSE.
	//THE FIRST RETURNS WHICH TYPE OF NODE WE SHOULD NEXT PICK TO SPLIT.
	//THE SECOND RETURNS, GIVEN A CERTAIN TYPE, WHICH TYPE ITS CHILDREN MUST HAVE.
	//THIS IS NOT NECESSARILY THE SAME (ALTHOUGH BY DEFAULT IT WILL BE).
	
	/**
	 * Given the type of the previous node that was split, returns the next node type.
	 * 
	 * By default it increases the type by 1, and returns 1 if the type already has the highest value.
	 * This method can be overridden however.
	 */
	protected int getNextTypeOfNodeToSplit(int lastNodeType) {	
		
		//if no node has been split before, then the first node to split will be the root node.
		if(lastNodeType == -1){
			return 0;
		}
		
		return getChildNodeType(lastNodeType);
	}
	
	/**
	 * Given a node type, it must return the type its children should have.
	 * 
	 * @param type
	 * @return
	 */
	protected abstract int getChildNodeType(int typeOfNodeToSplit);
	
	/*
	protected int getChildNodeType(int type){
		
		//0->1, 1->2, 2->3, 3->1  //After 3 comes 1 because we skip the root type.
		
		if(type == NUM_NODE_TYPES){
			return 1;
		}
		
		return type + 1;		
	}
	*/
	
	
	/**
	 * By default, if there are n types of nodes, then any node with type n-1 represents a deal.
	 * This method can be overridden however.
	 * @param node
	 * @return
	 */
	protected boolean canBeProposed(NB3Node node){
		return node.getType() == NUM_NODE_TYPES;
	}
	

	
	
	/**
	 * Must return the labels to be used for the children of the node to split
	 * @param nodeToSplit
	 * @return
	 */
	protected abstract List<Object> getSplitLabels(NB3Node nodeToSplit);

	
	/**
	 * Necessary in order to create a proposal object (which needs to know the pa in order to determine whether how
	 * many agents still need to accept).
	 * Also necessary in order to calculate the opponent utility for the negotiation strategy. And to determine
	 * whether a node is individually rational.
	 * 
	 * @param branch A list of nodes that represents a branch in the tree from the root node to a leaf node.
	 * @return
	 */
	public abstract NB3AgentSet getParticipatingAgents(ArrayList<NB3Node> branch);
	
	
	/*public NB3AgentSet getParticipatingAgentSet(ArrayList<NB3Node> branch){
		
		List<String> agentNames = getParticipatingAgents(branch);
		
		NB3AgentSet agentSet = new NB3AgentSet();
		for(String agentName : agentNames){
			int agentID = this.sessionInfo.getAgentID(agentName);
			agentSet.add(agentID);
		}
		
		return agentSet;
	}
	
	public abstract List<String> getParticipatingAgents(ArrayList<NB3Node> branch);
	*/
	
	
	
	/**
	 * Adds for each given label a new child node under the given node
	 * 
	 * @param node The node to be split.
	 * @param labels The labels that are going to be used to label the new child nodes.
	 */
	private void splitNode(NB3Node nodeToSplit, List<Object> labels){
		
		if(nodeToSplit == null){
			search_logger.logNoNodeToSplit();
			return;
		}
		if(labels == null || labels.size() == 0){		
			search_logger.logNoLabelToSplit(nodeToSplit);
			return;			
		}
		
		int nextType = getChildNodeType(nodeToSplit.type);
		
		ArrayList<NB3Node> branch = nodeToSplit.getBranch();
		
		//***Now create a new node for each label and add it to the tree.
		for(Object label : labels){
			
			//1. Add a new child to the node to split.	
			NB3Node newChild = nodeToSplit.addNewChild(nextType, label);
			
			//add it to the branch list.
			branch.add(newChild);
			NB3AgentSet pa = getParticipatingAgents(branch);
			
			//3. Calculate its bounds
			this.calculateBounds(newChild, branch, pa);
			
			//4. Calculate its expansion heuristic
			float eh = this.calculateExpansionHeuristic(newChild);
			newChild.setExpansionHeuristic(eh);
			
			//5. If the node represents a rational deal, then store it.			
			storeProposalIfRational(newChild, myAgentID, null);			
			
			
			//6. Put it on the queue
			theTree.queue.add(newChild);
				
			//7. Log it.
			search_logger.logNode(newChild);
			search_logger.logBounds(newChild, pa);
			if(canBeProposed(newChild)  && isRational(newChild, true)){
				search_logger.logln("RATIONAL");
			}
			
			//remove the child from the branch list again, so we can add the next child to it.
			branch.remove(branch.size()-1);
		}
		
		
		
	}

	
	
	/**
	 * Checks if node n corresponds to a rational deal.
	 * If so, converts it into an Nb3Proposal object and stores this in the list
	 * of found proposals.
	 * 
	 * @param node
	 */
	protected NB3Proposal storeProposalIfRational(NB3Node node, int proposer, String id) {
		
		
		//if the node cannot be proposed, return. 
		//otherwise there are four possible situations:
	
		//not rational & found by us		--> don't create a proposal
		//not rational & not found by us. 	--> create a proposal, store it in allProposals.
		//rational & found by us.     		--> create a proposal, store it in allProposals and foundByUs 
		//rational & not found by us		--> create a proposal, store it in allProposals and proposedToUs

				
		if(! canBeProposed(node)){
			return null;
		}
		
		boolean isFoundByUs = (proposer == myAgentID);
		boolean rational = isRational(node, isFoundByUs);
		
		if(!rational && isFoundByUs){
			return null;
		}
		
		
		//Create a proposal and store it in allProposals.
		int discoverTime = getCurrentTime();		
		NB3Proposal newProp = new NB3Proposal(node, discoverTime, proposer, sessionInfo); //a proposal is created with default id.
		if(id != null){ 			//if a specific id is given, we apply that one. (in case it is an incoming message, so that we make sure we use the same id as the agent who sent the message)
			newProp.setID(id);
		}
		allProposals.put(newProp.getID(), newProp);	
		
		
		if(rational){
			if(isFoundByUs){
				foundByUs.add(newProp);
				numFoundByUs++;
			}else{
				proposedToUs.add(newProp);
				numProposedToUs++;
			}
			
		}
				
		return newProp;
	}


	
	
	
	
	
	//***PART 2: HANDLE INCOMING MESSAGES	
	
	/**
	 * Handles incoming messages.
	 * 
	 * In case of a proposal message, it generates a new branch of nodes corresponding to the proposal-
	 * 
	 * In case of an acceptance message, it updates the offer values and remembers that the sender
	 * of the message has accepted the corresponding proposal.
	 * 
	 * @param msg
	 */
	public void handleIncomingMessages(NB3Message msg){
		
		switch(msg.getType()){		
		
		case PROPOSE:
			
			handleIncomingProposeMessage(msg);
			
			break;
		
		case ACCEPT:	
			
			handleIncomingAcceptMessage(msg);
			
			break;
			
		case REJECT:
			//REMARK: For now we don't send any REJECT_PROPOSAL messages, so we will also not receive any...
			
			//when we receive a reject message, the corresponding proposal can be in one of two lists:
			//feasibleDeals or awaitingConfirmation
						
			nego_logger.logln(getCurrentTime() + " Received a rejection of proposal " + msg.getProposalID());
			
			NB3Proposal rejectedProp = allProposals.get(msg.getProposalID());
			
			//Remove the deal from the list:			
			if(!foundByUs.remove(rejectedProp)){
					proposedByUs.remove(rejectedProp);
			}		
			//also we need to adapt the expansion heuristic here?			
		
			break;
			
		case ILLEGAL:	 //a proposal was illegal
			
			nego_logger.logln(getCurrentTime() + " Proposal " + msg.getProposalID() + " is inconsistent with earlier confirmed agreements.");
			
			//Remove the deal from the list of deals proposedByUs or proposedToUs
			NB3Proposal refusedProp = allProposals.get(msg.getProposalID());
			proposedByUs.remove(refusedProp);
			proposedToUs.remove(refusedProp);
			break;
			
		case CONFIRM:
			
			//log
			nego_logger.logSeparationLineHead();
			nego_logger.logln(getCurrentTime() + " ***Received a CONFIRMATION of proposal " + msg.getProposalID());
			
			
			
			
			NB3Proposal confirmedProp = allProposals.get(msg.getProposalID());						
			if(confirmedProp == null){
				confirmedProp = new NB3Proposal(msg, sessionInfo);
				//we need to create a new Proposal object because, if we are not participating in the
			    //deal, we don't have this deal on our list of standing deals. 
			}
			
			handleDealConfirmation(confirmedProp);
			
			
			
			
			//log
			nego_logger.logSeparationLineFoot();
			break;
		}
		
		
	}
	
	/**
	 * Converts the actions of an incoming message into a list of labels that can be added to the tree.
	 * 
	 * Note that one should make sure the returned labels appear in the correct order so that the first label can be added
	 * directly under the root node, and the second label can be added under the node of the first label, etcetera.
	 * 
	 * This method is important because incoming proposals can only be evalutated if they are properly added to the tree.
	 * 
	 * @param actions
	 * @return
	 */
	public abstract List<Object> actions2Labels(List<? extends NB3Action> actions);

	
	/**
	 * Converts NodeLabels into Actions.
	 * 
	 * Necessary in order to create an Nb3Proposal object from a node.
	 * The Nb3Proposal object needs the actions so that we can propose it to another agent, or we can
	 * execute it once it  is accepted by all participating agents.
	 * 
	 * 
	 * @param labels  The first element in this list is the label of the root node, and the last element is the label of the lowest node.
	 * @return
	 */
	public abstract List<? extends NB3Action> branch2actions(List<NB3Node> branch);
	
	
	/**
	 * Updates the offer level of the agent who proposes or accepts a proposal, and re-orders the search tree.
	 * @param newOfferVal
	 * @param idOfAgentToUpdate
	 */
	protected void updateOffers(float newOfferVal, int idOfAgentToUpdate){
		
		if(newOfferVal > theTree.offer[idOfAgentToUpdate]){
			theTree.offer[idOfAgentToUpdate] = newOfferVal;
			nego_logger.logln("updating offer level of: " + idOfAgentToUpdate);
		
			theTree.reorderQueue(idOfAgentToUpdate); //first recalculates the exp heur of every node in the open list and then re-orders the list.
		}
	}
	
	/**
	 * When a proposal is accepted by all its participating agents, this method is called
	 * to update the world state and to adapt the tree.
	 * 
	 */		
	protected void handleDealConfirmation(NB3Proposal confirmedProp){
		
		//Move the proposal from standing proposals to confirmed deal.
		proposedByUs.remove(confirmedProp);
		proposedToUs.remove(confirmedProp);		
		confirmed.add(confirmedProp);	
		
		
		if(confirmedProp.getParticipatingAgents().contains(myAgentID)){
			numCommitments++;
		}

		
		ArrayList<? extends NB3Action> actions = confirmedProp.getActions();
		List<Object> labels = actions2Labels(actions);
		
		
		nego_logger.logDealConfirmation(confirmedProp);
			
		//Update the current world state
		currentState.update(actions);
		
		//add the labels to the tree, as far as they aren't present yet.
		NB3Node newRoot = addLabelsToTree(labels);
		
		//now make the node that represents the confirmed proposal into the new root.
		theTree.resetRoot(newRoot);
		
		//TODO: cut away infeasible branches!
		
		
		foundByUs.clear();
		proposedToUs.clear();
		proposedByUs.clear();
		
	}
	
	/**
	 * This is the old implementation of Nb3_v1.2
	 */
	/*
	protected void handleDealConfirmation1_2(Nb3Proposal confirmedProp){
		
		
		//Update the current world state
		currentState.update(confirmedProp.getActions());

		initialize(endTime); //initialize a new search, but the endtime of negotiations stays the same.
		foundByUs = new PriorityQueue<Nb3Proposal>(1000);
		proposedToUs = new LinkedList<Nb3Proposal>();
		proposedByUs = new LinkedList<Nb3Proposal>();
		//myOwner.lastProposalOpponentSatisfaction = 0;
	}*/
	
	
	/**
	 * The default implementation creates one incomingProposalNode
	 * for the entire proposal and places it directly under the root.
	 * 
	 * Old code. Not removed, just in case we may want to use it again some day.
	 * 
	 */
	/*protected void handleIncomingProposeMessage1_2(Nb3Message msg){
		
		
		//when we receive a proposal, we first create an IncomingProposalNode
		//this node is then added to the tree and also converted into a Proposal object.
	
		float[] lb = new float[numAgents];
		float[] e = new float[numAgents];
		float[] ub = new float[numAgents];
		
		//get the path to the root:
		ArrayList<? extends Nb3Action> actions = msg.getActions();
		ArrayList<Nb3Label> labels = (ArrayList<Nb3Label>) actions2Labels(actions);  //we need to convert the actions to labels, cause the bound calculations only accept labels.
		
		//check if the incoming proposal is legal in the current world state
		if(!currentState.isLegal(actions)){
			nego_logger.logIllegalProposalReceived(myAgentName, msg);				
			return;
		}
		
		nego_logger.logProposalReceived(myAgentName, msg);	
		
		
		//calculate the bounds for this proposal
		for(int i=0; i<numAgents; i++){		
			lb[i] = theHeuristics.calculateLowerBound(i, labels, currentState);
			e[i] = theHeuristics.calculateIntermediateValue(i, labels, currentState);
			ub[i] = theHeuristics.calculateUpperBound(i, labels, currentState);
		
		}	
		
		//String proposer = msg.getConversationId().substring(0, 1); //it seems we don't need this variable. note that proposer is now an int
		//String proposalID = msg.getConversationId();  //this is moved to inside the constructor
		
		Nb3IncomingProposalNode ipn = new Nb3IncomingProposalNode(theTree, msg, labels, getRootLabel(), lb, e, ub);
		


		//and calculate the expansion heuristic!
		theHeuristics.calculateExpansionHeuristic(ipn);
		
		
		//if the expansion heuristic is positive, then put it on the queue
		if(ipn.getExpansionHeuristic() > 0){
			theTree.queue.add(ipn);
		}
		
		//also add the node to the tree
		theTree.addNode(ipn);
		
		//log the bounds
		nego_logger.logIncomingProposalNode(ipn);
		


		//***Create a Proposal object and store it.
		Nb3Proposal newProp = new Nb3Proposal(ipn);
		proposedToUs.add(newProp);
		numProposedToUs++;
		allProposals.put(msg.getConversationId() , newProp);
		int proposer = newProp.getProposer();
		updateOffers(newProp.e[proposer], proposer);	

	}*/
	
	/**
	 * New implementation, adds the incoming proposal as a complete branch to the tree.
	 * Therefore, it first checks whether a subset of the incoming proposal is already represented
	 * in the tree, and then adds the rest of the proposal under the lowest node that represents
	 * the common subset.
	 * 
	 * 
	 * @param msg
	 */
	protected void handleIncomingProposeMessage(NB3Message msg){
		
		//int[] lb = new int[numAgents];
		//int[] e = new int[numAgents];
		
		//get the path to the root:
		List<? extends NB3Action> actions = msg.getActions();
		
		//check if the incoming proposal is legal in the current world state
		if(!currentState.isLegal(actions)){
			nego_logger.logIllegalProposalReceived(myAgentName, msg);				
			return;
		}		
		
		nego_logger.logProposalReceived(myAgentName, msg);
		
		List<Object> labels = actions2Labels(actions);  //we need to convert the actions to labels, cause the bound calculations only accept labels.
		NB3Node incomingProposalNode = addLabelsToTree(labels);
				
		if(incomingProposalNode == null){
			return; //this happens when the deadline passes while adding the labels to the tree.
		}
		
		NB3AgentSet pa = getParticipatingAgents(incomingProposalNode.getBranch());
		
		nego_logger.logIncomingProposalNode(incomingProposalNode, pa);
		
		//nego_logger.logln("Nb3Algorithm.handleIncomingProposeMessage() bounds of incoming proposal:");
		//nego_logger.logBounds(incomingProposalNode);
		
		int proposerID = sessionInfo.getAgentID(msg.getSender());
		
		//if the node is rational, then store its corresponding proposal.
		NB3Proposal incomingProposal = storeProposalIfRational(incomingProposalNode, proposerID, msg.getProposalID());	
		
		//Debug code
		if(incomingProposal != null){
			nego_logger.logln("Still waiting for: " + incomingProposal.waitingFor);
		}
		
		//update the offer value
		updateOffers(incomingProposalNode.e[proposerID], proposerID);
		

		
		
	}
	
	
	protected void handleIncomingAcceptMessage(NB3Message msg){
		
		NB3Proposal acceptedProp = allProposals.get(msg.getProposalID());
		int accepter = sessionInfo.getAgentID(msg.getSender()); //the id of the agent that sent the accept message
			
		if(acceptedProp != null){
			acceptedProp.setAcceptedBy(accepter);
			updateOffers(acceptedProp.e[accepter], accepter);
			nego_logger.logIncomingAcceptance(acceptedProp, accepter);
		}else{
			nego_logger.logln(getCurrentTime() +" received accept message for unknown proposal " + msg.getProposalID());
		}
	}
	
	/**
	 * Adds a branch to the tree that contains the given labels.
	 * This branch might already be partially present, so we first check where we need to add new labels.
	 * The new branch represents an incoming proposal. 
	 *  
	 *  This method is called whenever we receive a proposal or when a deal is confirmed, and allows us
	 *  to incorporate the proposal or confirmed deal into the search tree.
	 *  
	 *  WARNING: this method requires that the labels are given in the proper order in which they can be
	 *  added to the root node.
	 *  
	 * @param branchLabels
	 * @return The lowest node in the branch that represents the given labels.
	 */
	NB3Node addLabelsToTree(List<Object> branchLabels){
		
		NB3Node currentNode = theTree.root;
		List<NB3Node> children;
		boolean matchingLabelExists;
		
		for(Object lab : branchLabels){
			
			children = currentNode.getChildren(); 
			
			//see if any of the children of the current node has a label that matches the current label.
			matchingLabelExists = false;
			for(NB3Node child : children){    //note that if currentNode has just been added, it has no children yet, so this loop will be skipped.
				if(lab.equals(child.label)){
					matchingLabelExists = true;			
					currentNode = child;
					break;					
				}
			}
			if(!matchingLabelExists){
				
				int nextType = getChildNodeType(currentNode.type);
				
				//if not, then add it
				NB3Node newChild = currentNode.addNewChild(nextType, lab);				
				
				ArrayList<NB3Node> branch = newChild.getBranch();
				NB3AgentSet pa = getParticipatingAgents(branch);
				
				calculateBounds(newChild, branch, pa);
				
				//calculate the expansion heuristic
				float eh = calculateExpansionHeuristic(newChild);
				newChild.setExpansionHeuristic(eh);
				
				//if the expansion heuristic is positive, then put it on the queue
				if(newChild.getExpansionHeuristic() > 0){
					theTree.queue.add(newChild);
				}
				
				search_logger.logNode(newChild);
				search_logger.logBounds(newChild, pa);
				if(canBeProposed(newChild) &&  isRational(newChild, false)){
					search_logger.logln("RATIONAL");
				}
				
				currentNode = newChild;
			}
					
		}
		
		//the last node in the branch represents the proposal made by the opponent.
		return currentNode; 
		
	}
	
	
	
	

	//***PART 3: ACCEPT OR PROPOSE 
	public void acceptOrPropose(){		
		
		//***Calculate the aspiration levels
		calculateOppAsp();
		calculateMyAsp();
		
		//***Get the best deal that we found (applies the aspiration levels to determine the 'best'. Returns null if no deal is selfish enough.)
		NB3Proposal bestFoundByUs = selectBestFoundByUs();
							
		//***Get the best acceptable deal proposed to us (applies the aspiration levels to determine the 'best'. Returns null if no deal is selfish enough.)
		NB3Proposal bestProposedToUs = selectBestProposedToUs();
		
		if(bestProposedToUs == null && bestFoundByUs == null){
			//this happens if no deal found and no deal proposed to us is selfish enough.
			nego_logger.logln("Decision: WAIT, at time " + getCurrentTime() + " Currently none of the deals found by me or proposed to me is selfish enough. my aspiration level: " + my_asp + " opponent aspiration level: " + opp_asp);
			return;
		}
		
		
		nego_logger.logBestProposals(bestFoundByUs, bestProposedToUs, my_asp, opp_asp);
	
		
		if(oursIsBetter(bestFoundByUs, bestProposedToUs)){
				
				propose(bestFoundByUs);  //Propose the deal!
				
				nego_logger.logln();
				nego_logger.logln("Nb3Algorithm.acceptOrPropose()");
				nego_logger.logProposalMade(bestFoundByUs, myAgentID, theTree.original_gub[myAgentID], my_asp, opp_asp);
				
				return;
				
		}else{
			
				accept(bestProposedToUs); //Accept the deal!
				
				nego_logger.logAcceptedProposal(bestProposedToUs);

				return;
			
		}
	}
	
	/**
	 * Compares two proposal, assuming that one of them was discovered by our search algorithm, while
	 * the other one was proposed to us by one of the other negotiators.
	 * 
	 * Returns true if the one found by our search algorithm is better than the one proposed to us.
	 * Returns false otherwise.
	 * 
	 * @param found
	 * @param proposed
	 * @return
	 */
	protected boolean oursIsBetter(NB3Proposal found, NB3Proposal proposed){
		
		if(found == null){
			return false;
		}
		if(proposed == null){			
			return true;			
		}

		//if one of the two is altruistic enough and the other isn't, then the one that is altruistic enough 
		//is the considered the 'best'.
		if(found.getOppNormalizedUtility() < opp_asp && proposed.getOppNormalizedUtility() > opp_asp){
			return false;
		}		
		
		//in all other cases we can simply compare them by selfishness.
		return found.getMyNormalizedUtility() > proposed.getMyNormalizedUtility();
		
					
	}
	
	
	
	/**
	 * Among all deals found by us and that are selfish enough and altruistic enough, 
	 * returns the one that is most selfish.
	 * 
	 * If there is no deal that is selfish enough and altruistic enough
	 * this returns the most altruistic one that is still selfish enough.
	 * 
	 * If no deal found by us is selfish enough, returns null.
	 * 
	 * The proposal returned is not removed from its list. It will only be removed once it is proposed. 
	 * 
	 * @return
	 */
	private NB3Proposal selectBestFoundByUs(){
		
		NB3Proposal mostSelfish = null;
		NB3Proposal mostAltruisticAndSelfishEnough = null;
		
		
		//if op_norm_util is lower than op_asp we haven't conceded enough
		//if my_norm_util is lower than my_asp we have conceded too much
		//if there is no deal for which both values are higher than their minimum, we prefer to wait instead 
			//of accepting or proposing any deal.
		
		//A = all deals that are selfish enough.
		//B = all deals that are altruistic enough 
		//if A \cap B non-empty, return the most selfish deal from this set.
		//if A is empty, return null
		//if B is empty, return the deal in A that is most altruistic.
		
			
		//implementation:
		// 1.get the most selfish deal. 
		//  if it is not selfish enough, return null
		//  else if it is altruistic enough, return it.
		// 2.loop over all deals found that are selfish enough
		//   and store the one that is most altruistic.
		// 3. return that one.
		
		//get the most selfish proposal			
		mostSelfish = foundByUs.peek();
		if(mostSelfish == null || mostSelfish.normalizedUtility[myAgentID] < my_asp){
			return null;
		}else if(mostSelfish.getOppNormalizedUtility() >= opp_asp){
			return mostSelfish;
		}
		
		
		double max_val = 0;
		for(NB3Proposal proposal : foundByUs){
			
			//skip those proposals that are not selfish enough
			if(proposal.normalizedUtility[myAgentID] < my_asp){
				continue;
			}
			
			//store the most altruistic proposal found so far.
			double op_norm_util = proposal.getOppNormalizedUtility();
			if(op_norm_util > max_val){
				max_val = op_norm_util;
				mostAltruisticAndSelfishEnough = proposal;
			}
			
		}
		
		return mostAltruisticAndSelfishEnough;
		

	}
	
	
	/**
	 * 
	 * Returns the most selfish proposal that was proposed to us.
	 * If this best proposal however is not rational it returns null.
	 * 
	 * The proposal returned is not removed from its list.
	 * 
	 * @return
	 */
	protected NB3Proposal selectBestProposedToUs(){
		
		//that is: select the proposal that was proposed to us with the highest value of e for us,
		//under the restriction that we haven't already accepted it, and that it is profitable to us.
		
		float max = 0;
		NB3Proposal bestProposedToUs = null;
		for(NB3Proposal prop : proposedToUs){
			if(!prop.isAcceptedByMe && prop.normalizedUtility[myAgentID] > max){
				max = prop.normalizedUtility[myAgentID];
				bestProposedToUs = prop;
			}			
		}
		
		if(max >= my_asp){
			return bestProposedToUs;
		}else{
			return null;
		}
		
		
	}
	
	/**
	 * Propose the given proposal to the other agents participating in it.
	 * 
	 * @param nextProposal
	 */
	protected void propose(NB3Proposal nextProposal){
		
		//Propose the deal, using a domain specific method.
		sendMessage(nextProposal, MsgType.PROPOSE);
		
		//Then do some NB3 specific bookkeeping:
		numProposedByUs++;		
		foundByUs.remove(nextProposal);
		proposedByUs.add(nextProposal);
		

	}
	
	
	protected void accept(NB3Proposal nextProposal){

		
		//Check if we haven't already accepted it before.
		if(nextProposal.isAcceptedByMe){
			return;
		}		
		nextProposal.isAcceptedByMe = true;
		nextProposal.setAcceptedBy(myAgentID);
		
		numAcceptedByUs++;
		
		//Then accept the deal, using a domain specific method.
		sendMessage(nextProposal, MsgType.ACCEPT);

	}
	
	/**
	 * Calculates the opponent's aspiration level.
	 */
	protected void calculateOppAsp(){
		
		double t = getCurrentTime();
		double t_d = sessionInfo.deadLine -  sessionInfo.startTime;
		double  z = opp_deadline_level; 
		double a = opp_concession_degree; 	  

		opp_asp = z / (Math.exp(-a) - 1);	
		opp_asp *= ( Math.exp(-a * t / t_d) - 1 );

	}
	
	/**
	 * Calculates the our aspiration level.
	 */
	protected void calculateMyAsp(){
		
		double t = getCurrentTime();
		double t_d = sessionInfo.deadLine -  sessionInfo.startTime;
		double  z = 1 - my_deadline_level; 
		double a = my_concession_degree; 
		
		my_asp = z / (Math.exp(-a) - 1);		
		my_asp *= ( Math.exp(-a * t / t_d) - 1 );
		
		my_asp = 1 - my_asp;
		
		
		/*
		if(my_asp  < theTree.getMyCurrentNormalizedUtility()){
			my_asp  = theTree.getMyCurrentNormalizedUtility();
		}
		*/
	}
	
	
	
	public void calculateBounds(NB3Node node, ArrayList<NB3Node> branch, NB3AgentSet pa){
		
		
		for(int agentID = 0; agentID<this.getAgentNames().length; agentID++){
			
			if(pa.contains(agentID) || node.parent == null){
				node.ub[agentID] = calculateUpperBound(agentID, branch, this.currentState);
				node.e[agentID] = calculateIntermediateValue(agentID, branch, this.currentState);
				node.lb[agentID] = calculateLowerBound(agentID, branch, this.currentState);
			}else{
				node.ub[agentID] = theTree.root.ub[agentID];
				node.e[agentID] = theTree.root.e[agentID];
				node.lb[agentID] = theTree.root.lb[agentID];
			}
		}
		
		//WARNING: when implementing the bound calculations below, it is often temping to simply copy the values from a parent node in 
		// case you know that the value hasn't changed. You should be careful with this however, because the values are only calculated
		// for agents that are in the participating agents. Therefore, the bound for a parent node may not have been set yet. 
		
	}
	
	public abstract float calculateUpperBound(int agentID, List<NB3Node> branch, NB3WorldState ws);
	
	public abstract float calculateIntermediateValue(int agentID, List<NB3Node> branch, NB3WorldState ws);
	
	public abstract float calculateLowerBound(int agentID, List<NB3Node> branchs,	NB3WorldState ws);

	protected abstract NB3Message convertMessage(Object domainMessage);
	
	protected abstract void sendMessage(NB3Proposal proposal, MsgType performative);
	
	
	
	/**
	 * Default, domain independent implementation of expansion heuristic.
	 * May be overridden with a domain-specific implementation. 
	 * 
	 * @param node
	 */
	protected float calculateExpansionHeuristic(NB3Node node) {
				
		float eh = 1;
		
		//1. First calculate the probability that all other participating agents will accept one nodes in the subtree under the given node.
		
		// get the set of agents participating in this deal. 
		NB3AgentSet pa = getParticipatingAgents(node.getBranch());
		
		
		//for each of these agents calculate the probability that it will accept the deal and take the product of these values.
		Arrays.fill(node.prob, -1);
		for(int ag : pa){		
						
			if(ag == myAgentID) {
				continue;
			}
			
			node.prob[ag] = calculateProbability(node, ag);
			eh *= node.prob[ag];
		}
		
		//2. Then multiply this value with our own current utility
		eh *= calculateExpectedUtility(node, myAgentID);
		
		return eh;
		
		
		
	}
	
	/**
	 * Calculates the probability that agent ag will accept the plan corresponding to the node.
	 * 
	 * We assume the following:
	 * rv < off.
	 * 
	 * if lb is bigger than off, the agent will accept for sure.
	 * if ub is lower than rv, the agent will certainly not accept.
	 * 
	 * We assume the probability that an agent will accept the deal increases as a function of the utility gain for that deal, linearly from 0 to 1 between rv and off.
	 * However, we don't know the exact utility that the agent will receive from the deal.
	 * But we do assume it has a probability distribution which is uniformly distributed between lb and ub.
	 *
	 */
	float calculateProbability(NB3Node node, int ag){
		

		float off = theTree.offer[ag];
		
		float rv = theTree.getRV(ag);
		
		float ub = node.ub[ag];
		float lb = node.lb[ag];
		
		
	
		//what if off == rv?
				

		
		
		//The likelihood that the agent will accept an offer increases linearly with utility from 0 to 1 between rv and off (function f)
		//The probability that the deal will have a certain utility is uniformly distributed between lb and ub. (function g)
		//We need the integral of the product of these two functions, divided by the total width ub-lb
		//The product f*g consists of a triangular-shaped part and a square-shaped part.
		
		
		//in the case that ub and lb are very close to each other, the calculations become too imprecise, so we have to calculate in an alternative way:
		if(Math.abs(ub-lb) < 0.01){
			if(lb>=off){   //if this node yields higher utility than the utility the agent demands, he will accept it. 
				return 1;
			}else if(ub<=rv){  //if his node yields lower utility than the reservation value, the agent will not accept it.
				return 0;
			}else{
				return ((ub-rv) / (off-rv));  //the probability the agent will accept is proportional to the fraction of e between rv and off.
			}
		}
		
		
		float triangle_underBound = Math.max(lb, rv);	// below lb g=0, below rv f=0.
		float triangle_upperBound = Math.min(ub,  off); // above off, f=1 (so the square area begins) above ub g=0.
		float square_underBound =   Math.max(off, lb);  // below off f!=1 (so we are in the triangle area), below lb g=0 
		float square_upperBound =  ub;  // above ub g=0
	

		
		float triangleArea;
		if(triangle_underBound >= triangle_upperBound){
			triangleArea = 0;
		}else{
			float triangleWidth = triangle_upperBound - triangle_underBound;
			float centerOfTriangleRange = (triangle_underBound + triangle_upperBound) / 2;
			float valueAtCenterOfTriangle = (centerOfTriangleRange - rv) / (off-rv);
			triangleArea = triangleWidth * valueAtCenterOfTriangle;
		}
		
		float squareArea;
		if(square_underBound >= square_upperBound){
			squareArea = 0;
		}else{
			squareArea = square_upperBound - square_underBound; //squareArea = 1*squareWidth
		}
		
		float prob = (squareArea + triangleArea) / (ub-lb);
		return prob;


	}
	
	/**
	 * The probability that agent ag will achieve positive utility can be estimated as:
	 * (ub - rv) / (ub - lb)
	 * 
	 * If the utility is positive, then the amount of utility that he expects to receive can be estimated as:
	 * (ub - rv)
	 * 
	 * Therefore, the expected utility is calculated as the product of these two values.
	 * 
	 * @param node
	 * @param ag
	 * @return
	 */
	float calculateExpectedUtility(NB3Node node, int ag){
		
		float retVal = (node.ub[ag] - node.theTree.getRV(ag));
		retVal *= retVal;
		retVal /= (node.ub[ag] - node.lb[ag]);
		
		return retVal;		
	}
	
	
	
	/**
	 * Recalculates the expansion heuristic when the offer value of one agent has changed.
	 * 
	 * @param node
	 * @param agent
	 */
	void reCalculateExpansionHeuristic(NB3Node node, int agent){
		
		if(agent == myAgentID){
			throw new ArithmeticException(); //if this happens, something is wrong.
		}		
		//remark: if the situation that n.prob[agent] == 0  occurs, something is wrong.

		//Divide the expansion heuristic by the old probability of adapted agent 
		  //and then multiply it with its new probability
		float eh = node.getExpansionHeuristic() / node.prob[agent]; 
		eh *=  calculateProbability(node, agent);
		node.setExpansionHeuristic(eh);		
	}
	


	
	//MISCELLANEOUS METHODS
	
	/**
	 * For nodes found by us: returns true if it is rational for me as well as for all other participating agents.
	 * For nodes corresponding to an incoming proposal:  returns true if it is rational for me.
	 * 
	 * @param node The node that represents a certain proposal.
	 * @param isFoundByUs 
	 * @return
	 */
	boolean isRational(NB3Node node, boolean isFoundByUs) {
		
		if(isFoundByUs){
			NB3AgentSet pa = getParticipatingAgents(node.getBranch());
			for(int ag : pa){			
				if(node.e[ag] - 0.1 <= theTree.getRV(ag)){
					return false;
				}
			}
		}
		
		if(node.e[myAgentID] - 0.1 <= theTree.getRV(myAgentID)){
			return false;
		}
		
		return true;

	}
		
	/**
	 * Write the contents of the loggers to a text file on hard disk.
	 */
	public void writeLogFiles(){
		this.search_logger.writeToFile();
		this.nego_logger.writeToFile();
	}
	
	/**
	 * Returns true if the deadline has not yet passed.
	 * @return
	 */
	boolean checkTime(){
		return System.currentTimeMillis() < this.sessionInfo.deadLine;
	}
	
	
	
	/**
	 * Returns the time in milliseconds since the negotiations have started.
	 * @return
	 */
	int getCurrentTime(){
		return (int) (System.currentTimeMillis() - this.sessionInfo.startTime);
	}

	/** Returns the time in milliseconds to go until the deadline. */
	int getTimeTillDeadline(){
		return (int) (this.sessionInfo.deadLine - System.currentTimeMillis());
	}
	
	public int getNumberOfNodesGenerated(){
		return numNodes;
	}
}
