package ddejonge.nb3;



import java.util.ArrayList;
import java.util.Collections;
import java.util.List;






public class NB3Node implements Comparable<NB3Node>{
//must be public, because the extending NB3Algorithm must implement methods that take a node as argument 
	
	//FIELDS	

	//The tree to which this node belongs.
	NB3Tree theTree; 
	
	//Each node has an id. Mainly for debugging purposes.
	int nodeId;
	
	//The depth of this node in the tree. The root node has depth 0, children of the root have depth 1, grand children of the root have depth 2, etcetera.
	int depth;
	
	NB3Node parent = null;
	ArrayList<NB3Node> children = new ArrayList<NB3Node>();
	boolean isPruned = false;
	
	NB3AgentSet participatingAgents = null;
	
	//The label is the object that carries domain-specific information.
	Object label;
	int type;
	
	//BOUNDS
	float[] ub;
	float[] e;
	float[] lb;
	
	/**
	 *  The utility is the difference between the value of this node and the reservation value.
	 *  The normalized utility is the utility, divided by the difference between the maximum utility and the reservation value.
	 *   (e - rv) / (original_gub - rv)  
	 */
	private float[] normalizedUtility;  			
	
	/** the estimated probability that agent ag will accept any node that will be created in the subtree under this node. */
	float[] prob;		
	
	/** product of normalized utilities of all other participating agents, always bigger than 0, if calculated.*/
	private float opponentUtility = -1; 			
	private float expansionHeuristic;
	
	
	//Fields should be initialized as much as possible in the algorithm, rather than inside this class definition,
	//because not every implementation uses all fields
	
	
	//CONSTRUCTORS
	
	/**
	 * Constructor for the root node.
	 * @param theTree
	 */
	NB3Node(NB3Tree theTree){
		
		int numAgents = theTree.numAgents;

		this.theTree = theTree;
		
		this.type = 0;
		this.label = null;
				
		ub = new float[numAgents];
		e = new float[numAgents];	
		lb = new float[numAgents];	
		prob = new float[numAgents];
		
		nodeId = theTree.theAlgorithm.numNodes;
		depth = 0;
		theTree.theAlgorithm.numNodes++;
	}
	
	
	/**
	 * Constructor for any node other than the root.
	 */
	private NB3Node(NB3Node parent, int type, Object label){
		this.parent = parent;
		this.type = type;
		this.label = label;
		
		this.depth = parent.depth + 1;
		this.theTree = parent.theTree;
		
		ub = new float[theTree.numAgents];
		e = new float[theTree.numAgents];	
		lb = new float[theTree.numAgents];	
		prob = new float[theTree.numAgents];
		
		
		//give the node an id (and increase numNodes so that the next node will receive a different id)
		this.nodeId = theTree.theAlgorithm.numNodes;		
		theTree.theAlgorithm.numNodes++;
	}
	
	
	//METHODS
	/**
	 * Adds a child to this node
	 * @param newChild
	 */
	NB3Node addNewChild(int type, Object label){
		NB3Node newChild = new NB3Node(this, type, label);
		children.add(newChild);
		
		return newChild;
	}

	

	public int getDepth(){
		return this.depth;
	}
	

	/**
	 * Returns the normalized utility of the agent on which this algorithm is running.
	 */
	float getMyNormalizedUtility(){
		return getNormalizedUtility(theTree.theAlgorithm.myAgentID);
	}
	
	/** Returns the normalized utility of the agent with the given ID.*/
	float getNormalizedUtility(int agentID){
		
		if(normalizedUtility == null){
			calculateNormalizedUtilities();
		}
		
		return normalizedUtility[agentID];
		
	}
	
	/**
	 * Returns the product of the normalized utilities of all opponents participating in the branch from the root to this node.
	 * 
	 * @return
	 */
	float getOpponentUtility(){
		
		if(opponentUtility != -1){
			return opponentUtility;
		}
		
		opponentUtility = 1;  
		
		if(participatingAgents == null){
			participatingAgents = this.theTree.theAlgorithm.getParticipatingAgents(this.getBranch());
		}
		
		for(int agentID : participatingAgents){
			
			// Skip myself.
			if(agentID == theTree.theAlgorithm.myAgentID){
				continue; 
			}
			
			// Calculate the opponent's normalized utility.
			float nu = getNormalizedUtility(agentID);
			
			//if x is smaller than 0, return 0
			if(nu < 0){
				opponentUtility = 0;
				return 0;
			}
			
			//multiply the total opponent utility with the normalized utility of the current agent.
			opponentUtility *= nu;
		}
		
		return opponentUtility;
	}
	
	private void calculateNormalizedUtilities(){
		
		float utility = -1;
		float range = -1;
		normalizedUtility = new float[e.length];
		
		for(int n=0; n<e.length; n++){
			utility = e[n] - theTree.original_rv[n];					//utility can be negative. If positive, then utility <= range.
			range = theTree.original_gub[n] - theTree.original_rv[n];   //range is always positive.
			
			//if range == 0 (implying also that utility == 0), the normalized utility should be zero,
			//otherwise we would start proposing deals for which our utility is equal to the reservation value.			
			if(utility >= 0 && range < 0.001){
				normalizedUtility[n] = 0;
			}else{
				normalizedUtility[n] = utility/range;
			}	
		}

	}

	
	public int getType(){
		return this.type;
	}
	
	public Object getLabel(){
		return this.label;
	}
	
	//PROPERTIES DERIVED FROM THE BOUNDS
	float utility(int n){
		return (e[n] - theTree.getRV(n));
	}
	
	float potential(int n){
		return (ub[n] - lb[n]);
	}

	
	
	//ACCESS METHODS
	public NB3Node getParent(){
		return parent;
	}
	
	List<NB3Node> getChildren(){
		return children;
	}
	
	/**
	 * Returns a list containing the nodes in the path from the root to this node (excluding the root itself)
	 * The first element is a direct child of the root node, and the last element is this node.
	 * @return
	 */
	public ArrayList<NB3Node> getBranch(){
		
		//getBranch() and addNodesToList() are separated, so that a new ArrayList only needs to be created once.
		
		
		ArrayList<NB3Node> branch = new ArrayList<NB3Node>(this.depth + 1); //make the list a bit bigger then necessary so we can add more nodes to it.
		addNodesToList(branch);
		
		//nodes are added bottom to top, so we need to reverse the list for the more intuitive top-to-bottom ordering.
		Collections.reverse(branch);
		
		return branch;
	}
	
	

	//adds the current node to the list and then passes the list on to the parent.
	private void addNodesToList(ArrayList<NB3Node> branch){
		
		//if this is the root node, then don't add it.
		if(this.parent == null){
			return;
		}
		
		//add this node to the list.
		branch.add(this);
		
		//call this method on its parent node.
		parent.addNodesToList(branch);
		
	}
	
	

	/**
	 * Returns a list containing the labels in the branch of this node (excluding the root).
	 * The first element is the label of a direct child of the root node, and the last element is the label of this node.
	 * @return
	 */
	ArrayList<Object> getBranchLabels(){
		
		//getBranchLabels() and addLabelsToPath() are separated, so that a new ArrayList only needs to be created once.
		
		ArrayList<Object> labels = new ArrayList<Object>(this.depth);
		addLabelsToList(labels);
		
		//nodes are added bottom to top, so we need to reverse the list for the more intuitive top-to-bottom ordering.
		Collections.reverse(labels);
		
		return labels;
	}
	
	
	private void addLabelsToList(ArrayList<Object> labels){
		
		//if this is the root node, then we have finished.
		if(this.parent == null){
			return;
		}
		
		//add the label of this node to the list.
		labels.add(this.label);
		
		//call this method on its parent node.
		parent.addLabelsToList(labels);
		
	}
	
	
	public float getUpperBound(int agentID){
		return this.ub[agentID];
	}
	
	public float intermediateValue(int agentID){
		return this.e[agentID];
	}
	
	public float getLowerBound(int agentID){
		return this.lb[agentID];
	}
	
	
	float getExpansionHeuristic(){
		return expansionHeuristic;
	}
	
	void setExpansionHeuristic(float eh){
		expansionHeuristic = eh;
	}
	

	
	//COMPARABLE
	/**
	 * Compares this node with another node, based on the expansion heuristic of the two nodes.
	 * This allows us to store the node in a priority queue.
	 */
	public int compareTo(NB3Node otherNode){
		
		//test which node has the highest expansion heuristic
		if(this.getExpansionHeuristic() > otherNode.getExpansionHeuristic()){
			return -1;
		}
		if(this.getExpansionHeuristic() < otherNode.getExpansionHeuristic()){
			return 1;
		}
		
		//if they are still equal, return 0
		return 0;
		
	}
	

	public String toString(){
		
		String s = this.label.toString();
		return s + " " + "exp: " + Math.round(this.getExpansionHeuristic() * 100);
	}
	


}
