package ddejonge.nb3;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.PriorityQueue;






/**
 * This class stores NB3Nodes sorted by their expansion heuristic, in several priority queues:
 * for each type of node a separate queue.
 * 
 * By default, the queue is a priority queue. The user can however override this.
 * The default capacity of the queue is 20,000
 * 
 * @author Dave de Jonge
 *
 */
class NB3NodeQueue implements Iterable<NB3Node> {
	
	
	
	//FIELDS
	static final int defaultCapacity = 20*1000;
	protected int numQueues;
	
	/**A list of priority queues; one priority queue for each node type.*/
	protected ArrayList<PriorityQueue<NB3Node>> queues; 
	
	//CONSTRUCTORS
	
	NB3NodeQueue(){
		this(1, defaultCapacity);
	}
	
	NB3NodeQueue(int numQueues){
		this(numQueues, defaultCapacity);
	}
	
	NB3NodeQueue(int numQueues, int  capacity){
		
		this.numQueues = numQueues;
		queues = new ArrayList<PriorityQueue<NB3Node>>(numQueues);
		for(int i=0; i<numQueues; i++){
			queues.add(new PriorityQueue<NB3Node>(capacity));
		}
		
		
	}
	
	//METHODS
	protected PriorityQueue<NB3Node> getQueue(int typeOfNodeToSplit){
		
		//note that the node type doesn't always have to coincide with the priority list to return, since there might be one queue for
		//several node types. In that case one can override this method.
		
		//default implementation: the number of the node type is exactly the number of the queue.		
		return queues.get(typeOfNodeToSplit);
	}
	
	
	void add(NB3Node n){		
		getQueue(n.getType()).add(n);
	}
	
	NB3Node poll(int typeOfNodeToSplit){
		return getQueue(typeOfNodeToSplit).poll();
	}
	
	NB3Node peek(int typeOfNodeToSplit){		
		return getQueue(typeOfNodeToSplit).peek();
	}

	/** Returns a string displaying the size of each priority queue. */
	String sizes(){
		
		String sizesString = "(";
		
		for(int i=0; i<numQueues-1; i++){
			sizesString += queues.get(i).size();
			sizesString += ", ";
		}
		sizesString += queues.get(numQueues-1).size();
		sizesString += ")";
		
		return sizesString;
		
	}
	
	/** Returns the total number of nodes stored 
	 *   i.e. the sum of the sizes of the priority queues.
	 */
	int size(){
		int size = 0;
		for(int i=0; i<numQueues; i++){
			size += queues.get(i).size();
		}
		return size;
	}
	
	
	/** 
	 * Calling this method ensures that all queues are sorted again.
	 * This is necessary if any agent accepts a proposal, because its offer value changes, and hence the expansion heuristic
	 *  of each node changes, which means that the queues are no longer sorted correctly.
	 * @param adaptedAgentIndex
	 */
	void reorder(int adaptedAgentIndex){
		//adaptedAgentIndex records which agent has changed its offer value. Not used for now, but might be by overriding methods.
		
		int s;
		for(int i=0; i<numQueues; i++){
			s = queues.get(i).size();
			if(s == 0){ 
				s++;
			}
			
			PriorityQueue<NB3Node> newQueue = new PriorityQueue<NB3Node>(s);
			newQueue.addAll(queues.get(i));	
			queues.set(i, newQueue);  //replace the old queue by the new queue.
			
		}
				
	}
	
	
	void clear(){
		for(PriorityQueue<NB3Node> q : queues){
			q.clear();
		}
	}
	
	
	
	//ITERATION
	@Override
	public Iterator<NB3Node> iterator() {
		
		ArrayList<Iterator<NB3Node>> iteratorList = new ArrayList<Iterator<NB3Node>>(numQueues);
		for(int i=0; i<numQueues; i++){
			iteratorList.add(queues.get(i).iterator());
		}
		
		return new Nb3NodeQueueIterator(iteratorList);
	}
	
}

class Nb3NodeQueueIterator implements Iterator<NB3Node>{
	
	
	ArrayList<Iterator<NB3Node>> iteratorList;
	
	//CONSTRUCTOR
	Nb3NodeQueueIterator(ArrayList<Iterator<NB3Node>> iteratorList){
		this.iteratorList =  iteratorList;
	}
	
	@Override
	public boolean hasNext(){
		
		for(int i=0; i<iteratorList.size(); i++){
			if(iteratorList.get(i).hasNext()){
				return true;
			}
		}
		return false;
		
	}
	
	@Override
	public NB3Node next(){
		
		for(int i=0; i<iteratorList.size(); i++){
			if(iteratorList.get(i).hasNext()){
				return iteratorList.get(i).next();
			}
		}
		return null;
		
	}

	public void remove() {
		throw new RuntimeException("Nb3NodeQueueIterator.remove() Error! remove() is not implemented.");
	}
	
}
