package ddejonge.nb3;



/**
 * The class that represents the NB3 search tree.
 * @author Dave de Jonge
 *
 */
class NB3Tree {
	
	/** The root node of the tree.*/
	public NB3Node root;
	
	/** The container to store the nodes in. */ 
	public NB3NodeQueue queue;  
	
	
	/**
	 * The original gub  (the global upper bound, i.e. the upper bound of the root node) is set equal to the upper bound of the root when the tree is initialized. 
	 *  After that it is not changed anymore. Note that the normal gub does change, whenever a deal is confirmed and the node corresponding to that deal becomes the new root.
	 */
	public float[] original_gub = null;
	
	/**
	 * The original rv  (the reservation value, i.e. the node value of the root node) is set equal to the node value of the root when the tree is initialized. 
	 *  After that it is not changed anymore. Note that the normal rv does change, whenever a deal is confirmed and the node corresponding to that deal becomes the new root.
	 */
	public float[] original_rv = null;
	
	/** The number of agents. */
	int numAgents;
	
	/** The lowest value that each agent has already accepted or proposed to receive (e.g. the lowest price that a salesman has offered to us).*/
	public float [] offer;  	 
	
	public NB3Algorithm theAlgorithm;
	

	//CONSTRUCTOR
	public NB3Tree(NB3Algorithm theAlgorithm, NB3NodeQueue queue, int numAgents){
		
		this.theAlgorithm = theAlgorithm; 		
		this.queue = queue;
		this.numAgents = numAgents;
		this.offer = new float[numAgents];
		
		//We must first set all fields before we pass this NB3Tree object to the constructor of the root node, because the constructor uses these fields.
		
		// Create the root node.
		root = new NB3Node(this);
		
		// Store the root node. 
		queue.add(root);
			
	}	
	

	
	
	
	
	/**
	 * Inverts the path from the new root to the original node
	 * @param newRoot
	 */
	/*
	public void resetRoot(Nb3Node newRoot){
		
		//iterative algorithm,
		//starts at the new root, and climbs up in the tree until it reaches the original root
		//for each node that is passed by the algorithm, the parent becomes a child, 
		//and the child that was handled in the previous iteration becomes the new parent.
		
		
		Nb3Node node = newRoot;   //the node handled by the current iteration. starting with the new root.
		Nb3Node newParent = null; //the new parent of the node handled by the current iteration. (i.e. the node handled by the previous iteration)
		Nb3Node oldParent;		  //the old parent of the node handled by the current iteration (i.e. the node handled by the next iteration)
		
		
		while(node != null){			
			
			oldParent = node.getParent();
			
			node.setParent(newParent);
			
			if(newParent != null) //is null at the first iteration, that is: when node is the new root.
				node.removeChild(newParent);
			
			if(oldParent != null) //is null at the last iteration, that is: when node is the original root.
				node.addChild(oldParent);
			
			newParent = node;
			node = oldParent;		
			
		}		

	}*/
	
	/**
	 * Makes the given node the new root.
	 * This method is called whenever a deal is confirmed. The node corresponding to that deal then becomes the new root node.
	 * 
	 * @param newRoot
	 */
	void resetRoot(NB3Node newRoot){
		
		this.root = newRoot;
		
		//TODO: prune all nodes that are not under root.
		
	}

	
	void prune(NB3Node node){
		
		node.isPruned = true;
		for(NB3Node child : node.children){
			prune(child);
		}
		
		if(node.parent != null && !node.parent.isPruned){
			prune(node.parent);
		}
		
	}
	

	float getGUB(int agentId){
		return this.root.ub[agentId];
	}
	
	float getRV(int agentId){
		return this.root.e[agentId];
	}
	
	float getGLB(int agentId){
		return this.root.lb[agentId];
	}
	
	
	
	
	
	/**
	 * Recalculates the expansion heuristic for every node in the queue.
	 * This is called whenever any agent makes a proposal or accepts a proposal, which changes his/her offer value.
	 * @param adaptedAgentIndex
	 */
	public void reorderQueue(int adaptedAgentIndex){

		//adaptedAgentIndex records which agent has changed its offer value.
	
		for(NB3Node n : queue){
			theAlgorithm.reCalculateExpansionHeuristic(n, adaptedAgentIndex);
		}
		
		queue.reorder(adaptedAgentIndex);
		
	}
	
	
	/**
	 * Returns our current normalized utility: the normalized utility that we have at the root node
	 * 	i.e. the normalized utility we receive when no deal is made. (maybe 'normalized reservation value' would be a better name)
	 * This value is important because we only want to accept proposals that have higher normalized utility than this
	 */
	public double getMyCurrentNormalizedUtility(){
		return getCurrentNormalizedUtility(theAlgorithm.myAgentID);
	}	
	
	/**
	 * Returns the current normalized utility of the agent with the given ID.
	 */
	public double getCurrentNormalizedUtility(int agentID){
		return root.getNormalizedUtility(agentID);
	}
	
	/** Remove all nodes from this tree.*/
	public void clear(){
		queue.clear();
	}
	
	
}
