package ddejonge.nb3;

import java.util.List;


/** Represents a possible state of the world.
 * 
 * This class is used mainly to calculate the heuristics.
 * We can distinguish three types of instantiations of this class:
 * 
 * 1) The initial state. This is the true state of the world in which the negotiations begin. It can be found inside the Nb3InstanceInfo class.
 * 2) The current state. During the negotiations the state can change as a result of actions being executed. Therefore the Nb3algorithm keeps a World State 
 *    variable that holds the current state of the world. Each time an agreement is made, the algorithm calls the update() method ito synchronize it with the real world.
 * 3) Hypothetical world states. Hypothetical world states are used to calculate the bounds of the nodes. The Nb3Algorithm does not explicitly use them. Instead,
 *    when a method of the Nb3Heuristics class is called, the current state is passed, together with the path to the root. Depending on the implementation, inside the heuristic function 
 *    a hypothetical world state might be created, or the bounds are calculated directly from the current state and the path.
 * 
 * @author Dave de Jonge
 *
 */
public abstract class NB3WorldState {

	int numAgents;
	
	protected NB3WorldState(int numAgents){
		this.numAgents = numAgents;
	}
	
	
	public int getNumAgents(){
		return numAgents;
	}
	
	
	public abstract boolean isLegal(List<? extends NB3Action> actions);

	
	/**
	 * This method is called when a deal is confirmed
	 * it lets the path act on the current world state
	 */
	public abstract void update(List<? extends NB3Action> arrayList);
	
	/**
	 * Makes a copy of this world state.
	 * Used by the Nb3Algorithm to instantiate the current state variable as a copy of the initial state variable.
	 */
	public abstract NB3WorldState copy();
	
}
